from django.conf import settings
from django.apps import apps
from django_sharding_library.constants import Backends
from django.utils.six import iteritems

from django_sharding_library.exceptions import NonExistentDatabaseException, ShardedModelInitializationException
from django_sharding_library.fields import ShardedIDFieldMixin, BasePostgresShardGeneratedIDField
from django_sharding_library.utils import register_migration_signal_for_model_receiver

PRE_MIGRATION_DISPATCH_UID = "PRE_MIGRATE_FOR_MODEL_%s"


def shard_storage_config(shard_group='default', shared_field='shard'):
    def configure(cls):
        setattr(cls, 'django_sharding__shard_group', shard_group)
        setattr(cls, 'django_sharding__shard_field', shared_field)
        setattr(cls, 'django_sharding__stores_shard', True)
        return cls
    return configure


def model_config(shard_group=None, database=None):
    """
    A decorator for marking a model as being either sharded or stored on a
    particular database. When sharding, it does some verification to ensure
    that the model is defined correctly.
    """
    def configure(cls):
        if database and shard_group:
            raise ShardedModelInitializationException('A model cannot be both sharded and stored on a particular database.')

        if not database and not shard_group:
            raise ShardedModelInitializationException('The model should be either sharded or stored on a database in the `model_config` decorator is used.')

        if database:
            if database not in settings.DATABASES or settings.DATABASES[database].get('PRIMARY'):
                raise NonExistentDatabaseException(
                    'Unable to place {} in {} as that is not an existing primary database in the system.'.format(cls._meta.model_name, database)
                )
            setattr(cls, 'django_sharding__database', database)

        postgres_shard_id_fields = list(filter(lambda field: issubclass(type(field), BasePostgresShardGeneratedIDField), cls._meta.fields))
        if postgres_shard_id_fields:
            database_dicts = [settings.DATABASES[database]] if database else [db_settings for db, db_settings in
                                                                              iteritems(settings.DATABASES) if
                                                                              db_settings["SHARD_GROUP"] == shard_group]
            if any([database_dict['ENGINE'] not in Backends.POSTGRES for database_dict in database_dicts]):
                raise ShardedModelInitializationException(
                    'You cannot use a PostgresShardGeneratedIDField on a non-Postgres database.')

            for field in postgres_shard_id_fields:
                register_migration_signal_for_model_receiver(apps.get_app_config(cls._meta.app_label),
                                                             field.migration_receiver,
                                                             dispatch_uid=PRE_MIGRATION_DISPATCH_UID % cls._meta.app_label)

        if shard_group:
            sharded_fields = list(filter(lambda field: issubclass(type(field), ShardedIDFieldMixin), cls._meta.fields))
            if not sharded_fields and not postgres_shard_id_fields:
                raise ShardedModelInitializationException('All sharded models require a ShardedIDFieldMixin or a '
                                                          'PostgresShardGeneratedIDField.')

            if not list(filter(lambda field: field == cls._meta.pk, sharded_fields + postgres_shard_id_fields)):
                raise ShardedModelInitializationException('All sharded models require a ShardedAutoIDField or '
                                                          'PostgresShardGeneratedIDField to be the primary key. Set '
                                                          'primary_key=True on the field.')

            if not callable(getattr(cls, 'get_shard', None)):
                raise ShardedModelInitializationException('You must define a get_shard method on the sharded model.')

            setattr(cls, 'django_sharding__shard_group', shard_group)
            setattr(cls, 'django_sharding__is_sharded', True)

        return cls
    return configure



from django.conf import settings
from django.db import models

from django_sharding_library.fields import BigAutoField


def _get_primary_shards():
    """
    Returns the names of databases which make up the shards and have no primary.
    """
    return sorted(
        filter(
            lambda db: not settings.DATABASES[db].get('PRIMARY', None) and settings.DATABASES[db].get('SHARD_GROUP', None),
            settings.DATABASES.keys()
        )
    )


class ShardedByMixin(models.Model):
    """
    This mixin is intended to be used with the included `save_shard_handler` in order
    to save the shard to the object. This is done by registering it on the User's pre-save signal.
    Look at the signal's docstring for additional information.

    Note: requires the use of `cls.shard_group`
    """
    django_sharding__shard_field = 'shard'
    django_sharding__stores_shard = True

    SHARD_CHOICES = ((i, i) for i in _get_primary_shards())

    shard = models.CharField(max_length=120, blank=True, null=True, choices=SHARD_CHOICES)

    class Meta:
        abstract = True


class TableStrategyModel(models.Model):
    """
    A model for use with the TableStrategy in order to generate ids using an
    autoincrimenting field. To use this, simply inherit from this class and give
    the model a more meaningful name e.g. InvoiceIDs.
    """
    id = BigAutoField(primary_key=True)
    stub = models.NullBooleanField(null=True, default=True, unique=True)

    class Meta:
        abstract = True


class ShardStorageModel(models.Model):
    """
    A model for storing shards.
    """
    SHARD_CHOICES = ((i, i) for i in _get_primary_shards())

    shard = models.CharField(max_length=120, choices=SHARD_CHOICES)
    shard_key = models.CharField(primary_key=True, max_length=120)

    class Meta:
        abstract = True


class ShardLookupQuerySet(models.QuerySet):
    def bulk_create(self, objs, batch_size=None):
        objs = list(objs)
        for obj in objs:
            if obj._state.db is None:
                obj._state.db = self._db

        return super().bulk_create(objs, batch_size)


class ShardLookupBaseModel(models.Model):
    """
    Unfortunatly when you call `model.objects.using(db).create(...) Django
    ignores the `db` passed to using and still tries to access the router.

    This can lead to many unnecesarry calls as well as circular resolution
    of the model.

    This is one way of solving this problem.
    """
    objects = ShardLookupQuerySet.as_manager()

    class Meta:
        abstract = True

    def save(self, **kwargs):
        if not self._state.db:
            self._state.db = kwargs["using"]

        return super().save(**kwargs)
